@@ -1,11 +1,12 @@
 from __future__ import annotations
-from typing import Any, Callable, Container, Dict, Iterable, Iterator, List, NoReturn, Optional, Reversible, Sequence, Set, Sized, TYPE_CHECKING, Tuple, Type, Generic, Union
+from typing import Any, Callable, Container, Deque, Dict, Iterable, Iterator, List, NoReturn, Optional, Reversible, Sequence, Set, Sized, TYPE_CHECKING, Tuple, Type, Generic, Union
 
 if TYPE_CHECKING:
     from .lookup import Lookup
     from .grouping import Grouping
     from .ordered_enumerable import OrderedEnumerable
 
+from .types_linq_error import InvalidOperationError, IndexOutOfRangeError
 from .more_typing import (
     TCollection,
     TDefault,
@@ -60,19 +61,25 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         if isinstance(index, int):
             # Sequence is an abstract base class without @runtime_checkable
             if not fallback and isinstance(iterable, Sequence):
-                # an appropriate implementation should raise IndexError
-                return iterable[index]
+                # an appropriate implementation should raise IndexError, or IndexOutOfRangeError
+                try:
+                    return iterable[index]
+                except IndexError as e:
+                    raise IndexOutOfRangeError(e)
             iterator = iter(iterable)
             try:
                 for _ in range(index):
                     next(iterator)
                 return next(iterator)
             except StopIteration:
-                raise IndexError('Not enough elements in the sequence')
+                raise IndexOutOfRangeError('Not enough elements in the sequence')
 
         else:  # isinstance(index, slice)
             if not fallback and isinstance(iterable, Sequence):
-                res = iterable[index]
+                try:
+                    res = iterable[index]
+                except IndexError as e:
+                    raise IndexOutOfRangeError(e)
                 return res if isinstance(res, Enumerable) else Enumerable(res)
             # we do not enumerate all values if the begin and the end only involve
             # nonnegative indices since in which case the sliced part can be obtained
@@ -89,19 +96,14 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                     yield from en.to_list()[s]
                     return
                 elif start_is_none:
-                    if s.stop >= 0:
-                        if step > 0:
-                            yield from en.take(s.stop)._every(step)
-                        else:
-                            yield from en.skip(s.stop + 1).reverse()._every(-step)
-                        return
-                    yield from en.to_list()[s]
-                elif s.start <= s.stop:
-                    en = en.skip(s.start).take(s.stop - s.start)
                     if step > 0:
-                        yield from en._every(step)
+                        yield from en.take(s.stop)._every(step)
                     else:
-                        yield from en.reverse()._every(-step)
+                        yield from en.skip(s.stop + 1).reverse()._every(-step)
+                    return
+                elif s.start <= s.stop and step > 0:
+                    yield from en.skip(s.start).take(s.stop - s.start)._every(step)
+                    return
                 elif step <= 0:
                     yield from en.skip(s.stop + 1).take(s.start - s.stop) \
                         .reverse()._every(-step)
@@ -138,7 +140,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
 
     @staticmethod
     def _raise_empty_sequence() -> NoReturn:
-        raise TypeError('Sequence is empty')
+        raise InvalidOperationError('Sequence is empty')
 
     def aggregate(self, *args) -> Any:
         if len(args) == 3:
@@ -281,7 +283,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
             default = args[0]
             try:
                 return self._getitem_impl(index, fallback=True)  # type: ignore
-            except IndexError:
+            except IndexOutOfRangeError:
                 return default
 
     @staticmethod
@@ -298,26 +300,30 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                 yield elem
         return Enumerable(inner)
 
+    @staticmethod
+    def _raise_no_such_element() -> NoReturn:
+        raise InvalidOperationError('No element satisfying condition')
+
     def first(self, *args: Callable[[TSource_co], bool]) -> TSource_co:
         if len(args) == 0:
             try:
                 return self.element_at(0)  # type: ignore
-            except IndexError as e:
-                raise ValueError(*e.args)
+            except IndexOutOfRangeError as e:
+                raise InvalidOperationError(e)
 
         else:  # len(args) == 1
             predicate = args[0]
             for elem in self:
                 if predicate(elem):
                     return elem
-            raise ValueError('No element satisfying condition')
+            self._raise_no_such_element()
 
     def first2(self, *args):
         if len(args) == 1:
             default = args[0]
             try:
                 return self.element_at(0)  # type: ignore
-            except IndexError:
+            except IndexOutOfRangeError:
                 return default
 
         else:  # len(args) == 2
@@ -398,7 +404,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
             for elem in self:
                 ret = elem
             if ret is _signal:
-                raise ValueError('Not enough elements in the sequence')
+                self._raise_empty_sequence()
 
         else:  # len(args) == 1
             predicate = args[0]
@@ -406,7 +412,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                 if predicate(elem):
                     ret = elem
             if ret is _signal:
-                raise ValueError('No element satisfying condition')
+                self._raise_no_such_element()
         return ret
 
     def last2(self, *args):
@@ -521,12 +527,16 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
             yield from self
         return Enumerable(inner)
 
+    @staticmethod
+    def _raise_count_negative() -> NoReturn:
+        raise InvalidOperationError('count must be nonnegative')
+
     @staticmethod
     def range(start: int, count: Optional[int]) -> Enumerable[int]:
         if count is not None:
             if count < 0:
-                raise ValueError('count must be nonnegative')
-            def inner(curr=start, cnt=count):
+                Enumerable._raise_count_negative()
+            def inner(curr=start, cnt=count):  # type: ignore
                 while cnt > 0:
                     yield curr
                     curr += 1
@@ -538,6 +548,21 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                     curr += 1
         return Enumerable(inner)
 
+    @staticmethod
+    def repeat(value: TResult, count: Optional[int] = None) -> Enumerable[TResult]:
+        if count is not None:
+            if count < 0:
+                Enumerable._raise_count_negative()
+            def inner(val=value, cnt=count):  # type: ignore
+                while cnt > 0:
+                    yield val
+                    cnt -= 1
+        else:
+            def inner(val=value):
+                while True:
+                    yield val
+        return Enumerable(inner)
+
     def reverse(self) -> Enumerable[TSource_co]:
         return Enumerable(lambda: self._reversed_impl(fallback=True))
 
@@ -581,6 +606,70 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                     yield result_selector(elem, sub)
         return Enumerable(inner)
 
+    def sequence_equal(self, second: Iterable[TSource_co]) -> bool:
+        me, she = iter(self), iter(second)
+        while True:
+            try:
+                lhs = next(me)
+            except StopIteration:
+                try:
+                    next(she)
+                    return False
+                except StopIteration:
+                    return True
+            try:
+                rhs = next(she)
+            except StopIteration:
+                return False
+            if not (lhs == rhs):
+                return False
+
+    def _find_single(self, res):
+        for elem in self:
+            if res is not _signal:
+                raise InvalidOperationError('Sequence does not contain exactly one element')
+            res = elem
+        return res
+
+    def _find_single_with_predicate(self, res, predicate):
+        for elem in self:
+            if predicate(elem):
+                if res is not _signal:
+                    raise InvalidOperationError(
+                        'There are multiple elements that satisfy condition: '
+                        f'{res} vs. {elem}'
+                    )
+                res = elem
+        return res
+
+    def single(self, *args: Callable[[TSource_co], bool]) -> TSource_co:
+        res: Any = _signal
+        if len(args) == 0:
+            res = self._find_single(res)
+            if res is _signal:
+                self._raise_empty_sequence()
+
+        else:  # len(args) == 1
+            predicate = args[0]
+            res = self._find_single_with_predicate(res, predicate)
+            if res is _signal:
+                self._raise_no_such_element()
+        return res
+
+    def single2(self, *args):
+        res: Any = _signal
+        if len(args) == 1:
+            default = args[0]
+            res = self._find_single(res)
+
+        else:  # len(args) == 2
+            predicate, default = args
+            res = self._find_single_with_predicate(res, predicate)
+        if res is _signal:
+            return default
+        return res
+
+
     def skip(self, count: int) -> Enumerable[TSource_co]:
         def inner():
             iterator = iter(self)
@@ -592,6 +681,70 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
             yield from iterator
         return Enumerable(inner)
 
+    def skip_last(self, count: int) -> Enumerable[TSource_co]:
+        if count <= 0:
+            return self.skip(0)
+        def inner():
+            iterator = iter(self)
+            q = Deque()
+            for elem in iterator:
+                if len(q) == count:
+                    while True:
+                        yield q.pop()
+                        q.appendleft(elem)
+                        try:
+                            elem = next(iterator)
+                        except StopIteration:
+                            break
+                else:
+                    q.appendleft(elem)
+        return Enumerable(inner)
+
+    def skip_while(self, predicate: Callable[[TSource_co], bool]) -> Enumerable[TSource_co]:
+        def inner():
+            iterator = iter(self)
+            for elem in iterator:
+                if not predicate(elem):
+                    yield elem
+                    break
+            yield from iterator
+        return Enumerable(inner)
+
+    def skip_while2(self, predicate: Callable[[TSource_co, int], bool]) -> Enumerable[TSource_co]:
+        def inner():
+            iterator = iter(self)
+            for i, elem in enumerate(iterator):
+                if not predicate(elem, i):
+                    yield elem
+                    break
+            yield from iterator
+        return Enumerable(inner)
+
+    def _sum_helper(self, selector, when_empty):
+        iterator = iter(self)
+        try:
+            sum_ = selector(next(iterator))
+        except StopIteration:
+            return when_empty()
+
+        for elem in iterator:
+            sum_ += selector(elem)
+        return sum_
+
+    def sum(self, *args) -> Any:
+        if len(args) == 0:
+            selector: Any = lambda x: x
+        else:  # len(args) == 1
+            selector = args[0]
+        return self._sum_helper(selector, lambda: 0)
+
+    def sum2(self, *args) -> Any:
+        if len(args) == 1:
+            selector, default = lambda x: x, args[0]
+        else: # len(args) == 2
+            selector, default = args
+        return self._sum_helper(selector, lambda: default)
+
     def take(self, count: int) -> Enumerable[TSource_co]:
         def inner():
             iterator = iter(self)
@@ -602,6 +755,45 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                 return
         return Enumerable(inner)
 
+    def take_last(self, count: int) -> Enumerable[TSource_co]:
+        if count <= 0:
+            return self.empty()
+        def inner():
+            iterator = iter(self)
+            try:
+                q = Deque((next(iterator),))
+            except StopIteration:
+                return
+            for elem in iterator:
+                if len(q) == count:
+                    while True:
+                        q.popleft()
+                        q.append(elem)
+                        try:
+                            elem = next(iterator)
+                        except StopIteration:
+                            break
+                else:
+                    q.append(elem)
+            yield from q
+        return Enumerable(inner)
+
+    def take_while(self, predicate: Callable[[TSource_co], bool]) -> Enumerable[TSource_co]:
+        def inner():
+            for elem in self:
+                if not predicate(elem):
+                    break
+                yield elem
+        return Enumerable(inner)
+
+    def take_while2(self, predicate: Callable[[TSource_co, int], bool]) -> Enumerable[TSource_co]:
+        def inner():
+            for i, elem in enumerate(self):
+                if not predicate(elem, i):
+                    break
+                yield elem
+        return Enumerable(inner)
+
     def to_dict(self,
         key_selector: Callable[[TSource_co], TKey],
         *args: Callable[[TSource_co], TValue],
